;   Emacs major mode for FAST
;   Copyright (C) 2003 Christophe Darlot, Alain Finkel (coordinator) and LSV,
;   CNRS UMR 8643 & ENS Cachan.
;
;   This file is part of FAST (Copyright (C) 2003 Jerome Leroux, Sebastien Bardin, Alain Finkel (coordinator) and LSV).
;
;   FAST is free software; you can redistribute it and/or modify
;   it under the terms of the GNU General Public License as published by
;   the Free Software Foundation; either version 2, or (at your option)
;   any later version.
;
;   FAST  is distributed in the hope that it will be useful,
;   but WITHOUT ANY WARRANTY; without even the implied warranty of
;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;   GNU General Public License for more details.
;
;   You should have received a copy of the GNU General Public License
;   along with FAST; see the file COPYING.  If not, write to
;   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

(defvar fast-mode-hook 'auto-insert-fast-template)
(defvar fast-mode-map nil
  "Keymap for FAST major mode.")

(if fast-mode-map 
    nil
  (setq fast-mode-map (make-sparse-keymap))
  (define-key fast-mode-map "\C-ct" 'insert-fast-transition)
  (define-key fast-mode-map "\C-c\C-f" 'run-fast)
  (define-key fast-mode-map "\C-c\C-t" 'run-fast-tc)
  (define-key fast-mode-map "\C-c\C-y" 'insert-fast-template-synchronisation)
  (define-key fast-mode-map "\C-c\C-s" 'insert-fast-template-strategy)
)

(fset 'show-fast-execution-buffer
   [?\C-x ?b ?F ?A ?S ?T ?  ?E ?x ?e ?c ?u ?t ?i ?o ?n return])

(defun run-fast ()
  (interactive)
  (save-buffer)
  (start-process "fast" "FAST Execution" "fast" (buffer-file-name))
  (execute-kbd-macro (symbol-function 'show-fast-execution-buffer)))

(defun run-fast-tc ()
  (interactive)
  (save-buffer)
  (start-process "fast" "FAST Execution" "fast" "-t" (buffer-file-name))
  (execute-kbd-macro (symbol-function 'show-fast-execution-buffer)))

(defun insert-fast-transition ()
  (interactive)
  (let ((nom (read-string "transition name: "))
	(from (read-string "from (state): "))
	(to (read-string "to (state): "))
	(guard (read-string "guard (presburger formula / press enter for \"true\"): "))
	(action (read-string "action: ")))
    (princ 'transition\ (get-buffer (buffer-name)))
    (princ nom (get-buffer (buffer-name)))
    (princ '\ :=\ { (get-buffer (buffer-name)))
    (terpri (get-buffer (buffer-name)))
    (princ '\ \ \ from\ :=\ (get-buffer (buffer-name)))
    (princ from (get-buffer (buffer-name)))
    (princ '\; (get-buffer (buffer-name)))
    (terpri (get-buffer (buffer-name)))
    (princ '\ \ \ to\ :=\  (get-buffer (buffer-name)))
    (princ to (get-buffer (buffer-name)))
    (princ '\; (get-buffer (buffer-name)))
    (terpri (get-buffer (buffer-name)))
    (princ '\ \ \ guard\ :=\ (get-buffer (buffer-name)))
    (if (equal guard "")
	(princ 'true (get-buffer (buffer-name)))
      (princ guard (get-buffer (buffer-name))))
    (princ '\; (get-buffer (buffer-name)))
    (terpri (get-buffer (buffer-name)))
    (princ '\ \ \ action\ :=\ (get-buffer (buffer-name)))
    (princ action (get-buffer (buffer-name)))
    (princ '\; (get-buffer (buffer-name)))
    (terpri (get-buffer (buffer-name)))
    (princ '}\; (get-buffer (buffer-name)))))

(defun auto-insert-fast-template ()
  (interactive)
  (if (file-exists-p buffer-file-name)
      (message "Opening an existing file")
    (if (yes-or-no-p "Do you want to insert a FAST file skeleton ? ")
	(let ((model (read-string "model name (default value: MyModel): "))
	      (strategy (read-string "strategy name (default value: MyStrat): "))
	      (maxstate (read-string "Set max size for Region (0 means infinity, default value: 2000) :"))
	      (maxacc (read-string "Set max of accelerations (0 means infinity, default value: 100) :")))
	  (princ 'model\ (get-buffer (buffer-name)))
	  (if (equal model "")
	      (princ 'MyModel (get-buffer (buffer-name)))
	    (princ model (get-buffer (buffer-name))))
	  (princ '\ { (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (princ 'var\ \; (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (princ 'states\ \; (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (princ '\/\*insert\ transitions\ here\*\/ (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (princ '} (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (princ 'strategy\ (get-buffer (buffer-name)))
	  (if (equal strategy "")
	      (princ 'MyStrategy (get-buffer (buffer-name)))
	    (princ strategy (get-buffer (buffer-name))))
	  (princ '\ { (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (princ 'setMaxState\( (get-buffer (buffer-name)))
	  (if (equal maxstate "")
	      (princ '2000 (get-buffer (buffer-name)))
	    (princ maxstate (get-buffer (buffer-name))))
	  (princ '\)\; (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (princ 'setMaxAcc\( (get-buffer (buffer-name)))
	  (if (equal maxacc "")
	      (princ '100 (get-buffer (buffer-name)))
	    (princ maxacc (get-buffer (buffer-name))))
	  (princ '\)\; (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (princ 'Region\ Init\ :=\ {}\; (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (princ 'Region\ Bad\ :=\ {}\; (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (princ 'Transitions\ t\ :=\ {}\; (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (princ 'Region\ Reach\ :=\ post*\(Init,t\)\; (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (princ 'if\ \(isEmpty\(Bad\ \&\&\ Reach\)\)\ then (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (princ 'print\(\"No\ configuration\ in\ the\ region\ bad\ is\ (get-buffer (buffer-name)))
	  (princ 'reachable\ from\ the\ region\ init\"\)\; (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (princ 'else (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (princ 'print\(\"A\ configuration\ in\ the\ region\ bad\ is\ (get-buffer (buffer-name)))
	  (princ 'reachable\ from\ the\ region\ init\"\)\; (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (princ 'endif (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name)))
	  (princ '} (get-buffer (buffer-name)))
	  (terpri (get-buffer (buffer-name))))
      (message "All right, use your fingers !"))))

(defun insert-fast-template-strategy ()
  (interactive)
  (princ 'strategy\ (get-buffer (buffer-name)))
  (princ 'MyStrategy (get-buffer (buffer-name)))
  (princ '\ { (get-buffer (buffer-name)))
  (terpri (get-buffer (buffer-name)))
  (terpri (get-buffer (buffer-name)))
  (princ 'setMaxState\( (get-buffer (buffer-name)))
  (princ '2000 (get-buffer (buffer-name)))
  (princ '\)\; (get-buffer (buffer-name)))
  (terpri (get-buffer (buffer-name)))
  (princ 'setMaxAcc\( (get-buffer (buffer-name)))
  (princ '100 (get-buffer (buffer-name)))
  (princ '\)\; (get-buffer (buffer-name)))
  (terpri (get-buffer (buffer-name)))
  (terpri (get-buffer (buffer-name)))
  (princ 'Region\ Init\ :=\ {}\; (get-buffer (buffer-name)))
  (terpri (get-buffer (buffer-name)))
  (terpri (get-buffer (buffer-name)))
  (princ 'Region\ Bad\ :=\ {}\; (get-buffer (buffer-name)))
  (terpri (get-buffer (buffer-name)))
  (terpri (get-buffer (buffer-name)))
  (princ 'Transitions\ t\ :=\ {}\; (get-buffer (buffer-name)))
  (terpri (get-buffer (buffer-name)))
  (terpri (get-buffer (buffer-name)))
  (princ 'Region\ Reach\ :=\ post*\(init,t\)\; (get-buffer (buffer-name)))
  (terpri (get-buffer (buffer-name)))
  (terpri (get-buffer (buffer-name)))
  (princ 'if\ \(isEmpty\(bad\ \&\&\ reach\)\)\ then (get-buffer (buffer-name)))
  (terpri (get-buffer (buffer-name)))
  (princ 'print\"No\ configuration\ in\ the\ region\ bad\ is\ (get-buffer (buffer-name)))
  (princ 'reachable\ from\ the\ region\ init\"\)\; (get-buffer (buffer-name)))
  (terpri (get-buffer (buffer-name)))
  (princ 'else (get-buffer (buffer-name)))
  (terpri (get-buffer (buffer-name)))
  (princ 'print\(\"A\ configuration\ in\ the\ region\ bad\ is\ (get-buffer (buffer-name)))
  (princ 'reachable\ from\ the\ region\ init\"\)\; (get-buffer (buffer-name)))
  (terpri (get-buffer (buffer-name)))
  (princ 'endif (get-buffer (buffer-name)))
  (terpri (get-buffer (buffer-name)))
  (princ '} (get-buffer (buffer-name)))
  (terpri (get-buffer (buffer-name))))

(defun insert-fast-template-synchronisation ()
  (interactive)
  (princ 'Synchronisation\ :=\ \ { (get-buffer (buffer-name)))
  (save-excursion 
    (princ '\ }\; (get-buffer (buffer-name)))
    (terpri (get-buffer (buffer-name)))))

(setq auto-mode-alist
      (append
       '(("\\.fst\\'" . fast-mode))
       auto-mode-alist))

(defconst fast-font-lock-keywords-1
  (list
   '("\\<\\(Synchronisation\\|model\\|strategy\\)\\>" . font-lock-type-face)
   '("\\<\\(states\\|transition\\|var\\)\\>" . font-lock-builtin-face)
   '("\\<\\(action\\|from\\|guard\\|to\\)\\>" . font-lock-variable-name-face))
  "Minimal highlighting expressions for FAST mode.")

(defconst fast-font-lock-keywords-2
  (append fast-font-lock-keywords-1
	  (list
	   '("\\<\\(Region\\|Transitions\\|e\\(?:lse\\|ndif\\)\\|if\\|p\\(?:ost\\|r\\(?:e\\|int\\)\\)\\|setMax\\(?:Acc\\|State\\)\\)\\>" . font-lock-keyword-face)
	   "Strategy highlighting in FAST mode.")))

(defconst fast-font-lock-keywords-3
  (append fast-font-lock-keywords-2
	  (list
	   '("\\<\\(False\\|True\\|and\\|e\\(?:qSet\\|xists\\)\\|f\\(?:alse\\|orall\\)\\|isEmpty\\|or\\|pre\\|s\\(?:tate\\|ubSet\\)\\|true\\)\\>" . font-lock-constant-face)))
  "Model highlighting in FAST mode.")

(defvar fast-font-lock-keywords fast-font-lock-keywords-3
  "Default highlighting expressions for FAST mode.")

(defun fast-indent-line ()
  "Indent current line as fast code."
  (interactive)
  (beginning-of-line)
  (if (bobp)
      (indent-line-to 0)	      ; First line is always non-indented
    (let ((not-indented t) cur-indent)
      (if (looking-at "^[. \t]*?}?;") ; If the line we are looking at is the end of a block, then decrease indentation
	  (progn
	    (save-excursion
	      (forward-line -1)
	      (setq cur-indent (- (current-indentation) 3)))
	    (if (< cur-indent 0)          ; We can't indent past the left margin
		(setq cur-indent 0)))
	(save-excursion
	  (while not-indented             ; Iterate backwards until we find an indentation hint
	    (forward-line -1)
	    (if (looking-at "^[. \t]*?}") ; This hint indicates that we need to indent at the level of the }
		(progn
		  (setq cur-indent (current-indentation))
		  (setq not-indented nil))
	      (if (looking-at "^[ \t]*\\(transition \\)") ; This hint indicates that we need to indent an extra level
		  (progn
		    (setq cur-indent (+ (current-indentation) 3)) ; Do the actual indenting
		    (setq not-indented nil))
		(if (bobp)
		    (setq not-indented nil)))))))
      (if cur-indent
	  (indent-line-to cur-indent)
	(indent-line-to 0))))) ; If we didn't see an indentation hint, then allow no indentation

(defvar fast-mode-syntax-table nil
  "Syntax table for fast-mode.")

(defun fast-create-syntax-table ()
  (if fast-mode-syntax-table
      ()
    (setq fast-mode-syntax-table (make-syntax-table))
    (modify-syntax-entry ?_ "w" fast-mode-syntax-table)
    (modify-syntax-entry ?/ ". 124b" fast-mode-syntax-table)
    (modify-syntax-entry ?* ". 23" fast-mode-syntax-table)
    (modify-syntax-entry ?\n "> b" fast-mode-syntax-table))
  (set-syntax-table fast-mode-syntax-table))

(defun fast-mode ()
  "Major mode for editing FAST files."
  (interactive)
  (kill-all-local-variables)
  (fast-create-syntax-table)
  ;; Set up font-lock
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults
	'(fast-font-lock-keywords))
  ;; Register our indentation function
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'fast-indent-line)
  (setq major-mode 'fast-mode)
  (setq mode-name "FAST")
  (use-local-map fast-mode-map)
  (run-hooks 'fast-mode-hook))

(provide 'fast-mode)

;;; fast-mode.el ends here
